unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, LCLType, Menus, Types, LazUTF8;

type

  TLBItem = class
    id: Integer;
    title: String;
    name: String;
    desc: String;
  end;

  { TForm1 }

  TForm1 = class(TForm)
    ListBox1: TListBox;
    MenuItem1: TMenuItem;
    MenuItem2: TMenuItem;
    PopupMenu1: TPopupMenu;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ListBox1DrawItem(Control: TWinControl; Index: Integer; ARect: TRect; State: TOwnerDrawState);
    procedure ListBox1MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure MenuItem1Click(Sender: TObject);
    procedure MenuItem2Click(Sender: TObject);
  private
    procedure FreeObjects;
  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }

procedure TForm1.ListBox1DrawItem(Control: TWinControl; Index: Integer; ARect: TRect; State: TOwnerDrawState);
var
  lb: TListBox;
  lbi: TLBItem;
  r: TRect;
  l, t, dotswidth: Integer;
  dots: String = '···';
begin
  if not (Control is TListBox) then
    exit;
  dotswidth := 36;
  lb := TListBox(Control);
  if odSelected in State then
    lb.Canvas.Brush.Color := $00FFDFBF
  else
    lb.Canvas.Brush.Color := clWhite;
  lb.Canvas.FillRect(ARect);
  r := ARect;
  r.Right := r.Right - dotswidth - 2;
  t := r.Top + 2;
  l := r.Left + 2;
  if Assigned(lb.Items.Objects[Index]) then
  begin
    lbi := TLBItem(lb.Items.Objects[Index]);
    //title
    lb.Canvas.Font.Size := 7;
    lb.Canvas.Font.Color := clNavy;
    lb.Canvas.Font.Style := [fsBold];
    lb.Canvas.TextRect(r, l, t, UTF8UpperCase(lbi.title + IntToStr(lbi.id)));
    t := t + 2 + lb.Canvas.GetTextHeight(UTF8UpperCase(lbi.title + IntToStr(lbi.id)));
    //name
    lb.Canvas.Font.Size := 9;
    lb.Canvas.Font.Color := clBlack;
    lb.Canvas.Font.Style := [];
    lb.Canvas.TextRect(r, l, t, lbi.name + IntToStr(lbi.id));
    t := t + 2 + lb.Canvas.GetTextHeight(lbi.name + IntToStr(lbi.id));
    //description
    lb.Canvas.Font.Size := 6;
    lb.Canvas.Font.Color := clGray;
    lb.Canvas.Font.Style := [fsItalic];
    lb.Canvas.TextRect(r, l, t, lbi.desc + IntToStr(lbi.id));
  end
  else
  begin
    lb.Canvas.Font.Size := 8;
    lb.Canvas.Font.Color := clBlack;
    lb.Canvas.Font.Style := [];
    lb.Canvas.TextRect(r, l, t, lb.Items[Index]);
  end;
  //dots
  r := ARect;
  r.Left := r.Right - dotswidth;
  lb.Canvas.Font.Size := 24;
  lb.Canvas.Font.Color := clBlack;
  t := ARect.Top + (ARect.Height - lb.Canvas.GetTextHeight(dots)) div 2;
  l := r.Left + (r.Width - lb.Canvas.GetTextWidth(dots)) div 2;
  lb.Canvas.TextRect(r, l, t, dots);
end;

procedure TForm1.ListBox1MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  p: TPoint;
  idx: Integer;
  r: TRect;
begin
  PopupMenu1.Tag := -1;
  if (Button = mbLeft) then
  begin
    p := Point(X, Y);
    idx := ListBox1.ItemAtPos(p, True);
    if (idx < 0) then
      exit;
    r := ListBox1.ItemRect(idx);
    r.Left := r.Right - 36;
    if not r.Contains(p) then
      exit;
    PopupMenu1.Tag := idx;
    PopupMenu1.PopUp;
  end;
end;

procedure TForm1.MenuItem1Click(Sender: TObject);
var
  lbi: TLBItem;
begin
  if (PopupMenu1.Tag >= 0) and (PopupMenu1.Tag < ListBox1.Items.Count) then
  begin
    if Assigned(ListBox1.Items.Objects[PopupMenu1.Tag]) then
    begin
      lbi := TLBItem(ListBox1.Items.Objects[PopupMenu1.Tag]);
      ShowMessage('Object item: ' + lbi.name + IntToStr(lbi.id));
    end
    else
      ShowMessage('String item: ' + ListBox1.Items[PopupMenu1.Tag]);
  end;
end;

procedure TForm1.MenuItem2Click(Sender: TObject);
var
  lbi: TLBItem;
begin
  if (PopupMenu1.Tag >= 0) and (PopupMenu1.Tag < ListBox1.Items.Count) then
  begin
    if Assigned(ListBox1.Items.Objects[PopupMenu1.Tag]) then
    begin
      lbi := TLBItem(ListBox1.Items.Objects[PopupMenu1.Tag]);
      lbi.id := lbi.id + Random(101);
      ListBox1.Items.Objects[PopupMenu1.Tag] := lbi;
      ListBox1.Invalidate;
    end
    else
      ShowMessage('Class item not assigned');
  end;
end;

procedure TForm1.FreeObjects;
var
  i: Integer;
  lbi: TLBItem;
begin
  for i := 0 to ListBox1.Items.Count - 1 do
  begin
    if Assigned(ListBox1.Items.Objects[i]) then
    begin
      lbi := TLBItem(ListBox1.Items.Objects[i]);
      ListBox1.Items.Objects[i] := nil;
      lbi.Free;
    end;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
var
  lbi: TLBItem;
  i: Integer;
begin
  Randomize;
  PopupMenu1.Tag := -1;
  //custom drawn
  ListBox1.Style := lbOwnerDrawFixed;
  ListBox1.ItemHeight := 50;
  for i := 0 to 10000 do
  begin
    lbi := TLBItem.Create;
    lbi.id := i;
    lbi.title := 'Title #';
    lbi.name := 'Item name No ';
    lbi.desc := 'Samle description for item no ';
    ListBox1.Items.AddObject(Format('Item name No %d', [i]), lbi);
    //ListBox1.Items.Add(Format('Item name No %d', [i]));
  end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FreeObjects;
end;

end.

